@@ -1,4 +1,4 @@
-/* ASN.1 object dumping code, copyright Peter Gutmann
+/* ASN.1 data display code, copyright Peter Gutmann 
    <pgut001@cs.auckland.ac.nz>, based on ASN.1 dump program by David Kemp, 
    with contributions from various people including Matthew Hamrick, Bruno 
    Couillard, Hallvard Furuseth, Geoff Thorpe, David Boyce, John Hughes, 
@@ -7,7 +7,7 @@
    Manger and several other people whose names I've misplaced.
 
    Available from http://www.cs.auckland.ac.nz/~pgut001/dumpasn1.c. Last
-   updated 21 May 2012 (version 20120521, if you prefer it that way).  
+   updated 8 June 2013 (version 20130608, if you prefer it that way).  
    To build under Windows, use 'cl /MD dumpasn1.c'.  To build on OS390 or 
    z/OS, use '/bin/c89 -D OS390 -o dumpasn1 dumpasn1.c'.
 
@@ -23,12 +23,12 @@
    errors, the fact that dumpasn1 will display an ASN.1 data item doesn't mean
    that the item is valid.
 
-   This version of dumpasn1 requires a config file dumpasn1.cfg to be present
-   in the same location as the program itself or in a standard directory where
-   binaries live (it will run without it but will display a warning message,
-   you can configure the path either by hardcoding it in or using an
-   environment variable as explained further down).  The config file is
-   available from http://www.cs.auckland.ac.nz/~pgut001/dumpasn1.cfg.
+   dumpasn1 requires a config file dumpasn1.cfg to be present in the same 
+   location as the program itself or in a standard directory where binaries 
+   live (it will run without it but will display a warning message, you can 
+   configure the path either by hardcoding it in or using an environment 
+   variable as explained further down).  The config file is available from 
+   http://www.cs.auckland.ac.nz/~pgut001/dumpasn1.cfg.
 
    This code assumes that the input data is binary, having come from a MIME-
    aware mailer or been piped through a decoding utility if the original
@@ -56,7 +56,7 @@
 
 /* The update string, printed as part of the help screen */
 
-#define UPDATE_STRING	"21 May 2012"
+#define UPDATE_STRING	"8 June 2013"
 
 /* Useful defines */
 
@@ -64,6 +64,9 @@
   #define FALSE	0
   #define TRUE	( !FALSE )
 #endif /* TRUE */
+#ifndef BYTE
+  typedef unsigned char		BYTE;
+#endif /* BYTE */
 
 /* Tandem Guardian NonStop Kernel options */
 
@@ -105,6 +108,7 @@
 #if ( defined( _WINDOWS ) || defined( WIN32 ) || defined( _WIN32 ) || \
 	  defined( __WIN32__ ) )
   #include <windows.h>
+  #include <io.h>					/* For _setmode() */
   #include <fcntl.h>				/* For _setmode() codes */
   #ifndef _O_U16TEXT 
 	#define _O_U16TEXT		0x20000	/* _setmode() code */
@@ -157,7 +161,7 @@
 
 /* Macros to avoid problems with sign extension */
 
-#define byteToInt( x )	( ( unsigned char ) ( x ) )
+#define byteToInt( x )	( ( BYTE ) ( x ) )
 
 /* The level of recursion can get scary for deeply-nested structures so we
    use a larger-than-normal stack under DOS */
@@ -248,8 +252,9 @@ typedef struct {
 	int tag;					/* Tag */
 	long length;				/* Data length */
 	int indefinite;				/* Item has indefinite length */
+	int nonCanonical;			/* Non-canonical length encoding used */
+	BYTE header[ 16 ];			/* Tag+length data */
 	int headerSize;				/* Size of tag+length */
-	unsigned char header[ 16 ];	/* Tag+length data */
 	} ASN1_ITEM;
 
 /* Configuration options */
@@ -341,7 +346,7 @@ static FILE *output;				/* Output stream */
 
 typedef struct tagOIDINFO {
 	struct tagOIDINFO *next;		/* Next item in list */
-	unsigned char oid[ MAX_OID_SIZE ];
+	BYTE oid[ MAX_OID_SIZE ];
 	int oidLength;
 	char *comment, *description;	/* Name, rank, serial number */
 	int warn;						/* Whether to warn if OID encountered */
@@ -372,11 +377,9 @@ static const char *configPaths[] = {
 #elif defined( __WIN32__ )
 
 static const char *configPaths[] = {
-	/* Windoze absolute paths.  Usually things are on C:, but older NT setups
-	   are easier to do on D: if the initial copy is done to C: (yeah, this
-	   code has been around for awhile, why do you ask?) */
-	"c:\\dos\\", "d:\\dos\\", "c:\\windows\\", "d:\\windows\\",
-	"c:\\winnt\\", "d:\\winnt\\",
+	/* Windoze absolute paths (yeah, this code has been around for awhile, 
+	   why do you ask?) */
+	"c:\\windows\\", "c:\\winnt\\", 
 
 	/* It's my program, I'm allowed to hardcode in strange paths that no-one
 	   else uses */
@@ -506,14 +509,18 @@ static char *idstr( const int tagID )
 
 /* Return information on an object identifier */
 
-static OIDINFO *getOIDinfo( char *oid, const int oidLength )
+static OIDINFO *getOIDinfo( const BYTE *oid, const int oidLength )
 	{
+	const BYTE oidByte = oid[ 1 ];
 	OIDINFO *oidPtr;
 
 	for( oidPtr = oidList; oidPtr != NULL; oidPtr = oidPtr->next )
 		{
-		if( oidLength == oidPtr->oidLength - 2 && \
-			!memcmp( oidPtr->oid + 2, oid, oidLength ) )
+		if( oidLength != oidPtr->oidLength - 2 )
+			continue;	/* Quick-reject check */
+		if( oidByte != oidPtr->oid[ 2 + 1 ] )
+			continue;	/* Quick-reject check */
+		if( !memcmp( oidPtr->oid + 2, oid, oidLength ) )
 			return( oidPtr );
 		}
 
@@ -656,7 +663,7 @@ static int readLine( FILE *file, char *buffer )
 		while( !feof( file ) )
 			{
 			/* Keep going until we hit the true EOF (or some sort of error) */
-			ch = getc( file );
+			( void ) getc( file );
 			}
 		}
 
@@ -667,8 +674,9 @@ static int readLine( FILE *file, char *buffer )
 
 static int processOID( OIDINFO *oidInfo, char *string )
 	{
-	unsigned char binaryOID[ MAX_OID_SIZE ];
-	int firstValue, value, valueIndex = 0, oidIndex = 3;
+	BYTE binaryOID[ MAX_OID_SIZE ];
+	long value;
+	int firstValue, valueIndex = 0, oidIndex = 3;
 
 	memset( binaryOID, 0, MAX_OID_SIZE );
 	binaryOID[ 0 ] = OID;
@@ -680,7 +688,7 @@ static int processOID( OIDINFO *oidInfo, char *string )
 					lineNo );
 			return( FALSE );
 			}
-		if( sscanf( string, "%d", &value ) != 1 || value < 0 )
+		if( sscanf( string, "%8ld", &value ) != 1 || value < 0 )
 			{
 			printf( "Invalid value in config file line %d.\n", lineNo );
 			return( FALSE );
@@ -702,7 +710,7 @@ static int processOID( OIDINFO *oidInfo, char *string )
 							lineNo );
 					return( FALSE );
 					}
-				binaryOID[ 2 ] = ( firstValue * 40 ) + value;
+				binaryOID[ 2 ] = ( firstValue * 40 ) + ( int ) value;
 				valueIndex++;
 				}
 			else
@@ -711,22 +719,22 @@ static int processOID( OIDINFO *oidInfo, char *string )
 
 				if( value >= 0x200000L )					/* 2^21 */
 					{
-					binaryOID[ oidIndex++ ] = 0x80 | ( value >> 21 );
+					binaryOID[ oidIndex++ ] = 0x80 | ( int ) ( value >> 21 );
 					value %= 0x200000L;
 					hasHighBits = TRUE;
 					}
 				if( ( value >= 0x4000 ) || hasHighBits )	/* 2^14 */
 					{
-					binaryOID[ oidIndex++ ] = 0x80 | ( value >> 14 );
+					binaryOID[ oidIndex++ ] = 0x80 | ( int ) ( value >> 14 );
 					value %= 0x4000;
 					hasHighBits = TRUE;
 					}
 				if( ( value >= 0x80 ) || hasHighBits )		/* 2^7 */
 					{
-					binaryOID[ oidIndex++ ] = 0x80 | ( value >> 7 );
+					binaryOID[ oidIndex++ ] = 0x80 | ( int ) ( value >> 7 );
 					value %= 128;
 					}
-				binaryOID[ oidIndex++ ] = value;
+				binaryOID[ oidIndex++ ] = ( int ) value;
 				}
 			}
 		while( *string && isdigit( byteToInt( *string ) ) )
@@ -750,7 +758,7 @@ static int processHexOID( OIDINFO *oidInfo, char *string )
 
 	while( *string && index < MAX_OID_SIZE - 1 )
 		{
-		if( sscanf( string, "%x", &value ) != 1 || value > 255 )
+		if( sscanf( string, "%4x", &value ) != 1 || value > 255 )
 			{
 			printf( "Invalid hex value in config file line %d.\n", lineNo );
 			return( FALSE );
@@ -857,6 +865,25 @@ static int readConfig( const char *path, const int isDefaultConfig )
 				if( !processOID( oidPtr, buffer + 6 ) )
 					return( FALSE );
 				}
+
+			/* Check that this OID isn't already present in the OID list.  
+			   This is a quick-and-dirty n^2 algorithm so it's not enabled 
+			   by default */
+#if 0
+			{
+			OIDINFO *oidCursor;
+
+			for( oidCursor = oidList; oidCursor->next != NULL; oidCursor = oidCursor->next )
+				{
+				if( oidCursor->oidLength == oidPtr->oidLength && \
+					!memcmp( oidCursor->oid, oidPtr->oid, oidCursor->oidLength ) )
+					{
+					printf( "Duplicate OID '%s' at line %d.\n",
+							buffer, lineNo );
+					}
+				}
+			}
+#endif /* 0 */
 			}
 		else if( !strncmp( buffer, "Description = ", 14 ) )
 			{
@@ -1173,7 +1200,8 @@ static void doIndent( const int level )
 
 /* Complain about an error in the ASN.1 object */
 
-static void complain( const char *message, const int level )
+static void complain( const char *message, const int messageParam, 
+					  const int level )
 	{
 	if( level < maxNestLevel )
 		{
@@ -1181,7 +1209,66 @@ static void complain( const char *message, const int level )
 			fprintf( output, INDENT_STRING );
 		doIndent( level + 1 );
 		}
-	fprintf( output, "Error: %s.\n", message );
+	fputs( "Error: ", output );
+	fprintf( output, message, messageParam );
+	fputs( ".\n", output );
+	noErrors++;
+	}
+
+static void complainLength( const ASN1_ITEM *item, const int level )
+	{
+#if 0
+	/* This is a general error so we don't indent the message to the level
+	   of the item */
+#else
+	if( level < maxNestLevel )
+		{
+		if( !doPure )
+			fprintf( output, INDENT_STRING );
+		doIndent( level + 1 );
+		}
+#endif /* 0 */
+	fprintf( output, "Error: %s has invalid length %d.\n", 
+			 idstr( item->tag ), item->length );
+	noErrors++;
+	}
+
+static void complainLengthCanonical( const ASN1_ITEM *item, const int level )
+	{
+	int i;
+
+#if 0
+	/* This is a general error so we don't indent the message to the level
+	   of the item */
+#else
+	if( level < maxNestLevel )
+		{
+		if( !doPure )
+			fprintf( output, INDENT_STRING );
+		doIndent( level + 1 );
+		}
+#endif /* 0 */
+	fputs( "Error: Length '", output );
+	for( i = item->nonCanonical; i < item->headerSize; i++ )
+		{
+		fprintf( output, "%02X", item->header[ i ] );
+		if( i < item->headerSize - 1 )
+			fputc( ' ', output );
+		}
+	fputs( "' has non-canonical encoding.\n", output );
+	noErrors++;
+	}
+
+static void complainInt( const BYTE *intValue, const int level )
+	{
+	if( level < maxNestLevel )
+		{
+		if( !doPure )
+			fprintf( output, INDENT_STRING );
+		doIndent( level + 1 );
+		}
+	fprintf( output, "Error: Integer '%02X %02X ...' has non-DER encoding.\n", 
+			 intValue[ 0 ], intValue[ 1 ] );
 	noErrors++;
 	}
 
@@ -1246,15 +1333,24 @@ static int displayUnicode( const wchar_t wCh, const int level )
 	   it's a one-way function, once we've set it a particular way we can't 
 	   go back (exactly what level of braindamage it takes to have an 
 	   implementation function like this is a mystery).  Other sources 
-	   suggest using setlocale() tricks, printf() with "%lc" as the format 
-	   specifier, and others, but this doesn't seem to work either */
+	   suggest using setlocale() tricks, printf() with "%lc" or "%ls" as the 
+	   format specifier, and others, but none of these seem to work properly 
+	   either */
 	if( level < maxNestLevel )
 		{
 #if 0
 		setlocale( LC_ALL, "" );
 		fputwc( wCh, output );
-#elif 0
+#elif 1
+		/* This (and the "%ls" variant below) seem to be the least broken 
+		   options */
 		fprintf( output, "%lc", wCh );
+#elif 0
+		wchar_t wChString[ 2 ];
+
+		wChString[ 0 ] = wCh;
+		wChString[ 1 ] = 0;
+		fprintf( output, "%ls", wChString );
 #else
 		if( fwide( output, 1 ) > 0 )
 			{
@@ -1285,8 +1381,10 @@ static int displayUnicode( const wchar_t wCh, const int level )
 
 /* Display an integer value */
 
-static void printValue( FILE *inFile, const int valueLength, const int level )
+static void printValue( FILE *inFile, const int valueLength, 
+					    const int level )
 	{
+	BYTE intBuffer[ 2 ];
 	long value;
 	int warnNegative = FALSE, warnNonDER = FALSE, i;
 
@@ -1304,6 +1402,11 @@ static void printValue( FILE *inFile, const int valueLength, const int level )
 				warnNonDER = TRUE;
 			if( ( value == 0xFF ) && ( ( ch & 0x80 ) == 0x80 ) )
 				warnNonDER = TRUE;
+			if( warnNonDER )
+				{
+				intBuffer[ 0 ] = ( int ) value;
+				intBuffer[ 1 ] = ch;
+				}
 			}
 		value = ( value << 8 ) | ch;
 		}
@@ -1312,17 +1415,19 @@ static void printValue( FILE *inFile, const int valueLength, const int level )
 	/* Display the integer value and any associated warnings */
 	printString( level, " %ld\n", value );
 	if( warnNonDER )
-		complain( "Integer has non-DER encoding", level );
+		complainInt( intBuffer, level );
 	if( warnNegative )
-		complain( "Integer has a negative value", level );
+		complain( "Integer has a negative value", 0, level );
 	}
 
 /* Dump data as a string of hex digits up to a maximum of 128 bytes */
 
-static void dumpHex( FILE *inFile, long length, int level, int isInteger )
+static void dumpHex( FILE *inFile, long length, int level, 
+					 const int isInteger )
 	{
 	const int lineLength = ( dumpText ) ? 8 : 16;
 	const int displayHeaderLength = ( ( doPure ) ? 0 : INDENT_SIZE ) + 2;
+	BYTE intBuffer[ 2 ];
 	char printable[ 9 ];
 	long noBytes = length;
 	int warnPadding = FALSE, warnNegative = isInteger, singleLine = FALSE;
@@ -1392,6 +1497,11 @@ static void dumpHex( FILE *inFile, long length, int level, int isInteger )
 				warnPadding = TRUE;
 			if( ( prevCh == 0xFF ) && ( ( ch & 0x80 ) == 0x80 ) )
 				warnPadding = TRUE;
+			if( warnPadding )
+				{
+				intBuffer[ 0 ] = prevCh;
+				intBuffer[ 1 ] = ch;
+				}
 			}
 		}
 	if( dumpText )
@@ -1429,25 +1539,25 @@ static void dumpHex( FILE *inFile, long length, int level, int isInteger )
 	if( isInteger )
 		{
 		if( warnPadding )
-			complain( "Integer has non-DER encoding", level );
+			complainInt( intBuffer, level );
 		if( warnNegative )
-			complain( "Integer has a negative value", level );
+			complain( "Integer has a negative value", 0, level );
 		}
 	}
 
 /* Convert a binary OID to its string equivalent */
 
 static int oidToString( char *textOID, int *textOIDlength,
-						const unsigned char *oid, const int oidLength )
+						const BYTE *oid, const int oidLength )
 	{
-	unsigned char uuidBuffer[ 32 ];
+	BYTE uuidBuffer[ 32 ];
 	long value;
 	int length = 0, uuidBufPos = -1, uuidBitCount = 5, i;
 	int validEncoding = TRUE, isUUID = FALSE;
 
 	for( i = 0, value = 0; i < oidLength; i++ )
 		{
-		const unsigned char data = oid[ i ];
+		const BYTE data = oid[ i ];
 		const long valTmp = value << 7;
 
 		/* Pick apart the encoding.  We keep going after hitting an encoding
@@ -1584,30 +1694,45 @@ static void dumpBitString( FILE *inFile, const int length, const int unused,
 	char *errorStr = NULL;
 
 	if( unused < 0 || unused > 7 )
-		complain( "Invalid number of unused bits", level );
+		complain( "Invalid number %d of unused bits", unused, level );
 	noBits = ( length * 8 ) - unused;
 
 	/* ASN.1 bitstrings start at bit 0, so we need to reverse the order of
 	   the bits if necessary */
-	if( length )
+	if( length > 0 )
 		{
 		bitString = fgetc( inFile );
+		if( bitString == EOF )
+			errorStr = "Truncated BIT STRING data";
 		fPos++;
 		}
 	for( i = noBits - 8; i > 0; i -= 8 )
 		{
-		bitString = ( bitString << 8 ) | fgetc( inFile );
+		const int ch = fgetc( inFile );
+
+		if( ch == EOF )
+			{
+			errorStr = "Truncated BIT STRING data";
+			break;
+			}
+		bitString = ( bitString << 8 ) | ch;
 		currentBitMask <<= 8;
 		remainderMask = ( remainderMask << 8 ) | 0xFF;
 		fPos++;
 		}
+	if( errorStr != NULL )
+		{
+		printString( level, "%c", '\n' );
+		complain( errorStr, 0, level );
+		return;
+		}
 	if( reverseBitString )
 		{
 		for( i = 0, bitFlag = 1; i < noBits; i++ )
 			{
 			if( bitString & currentBitMask )
 				value |= bitFlag;
-			if( !( bitString & remainderMask ) )
+			if( !( bitString & remainderMask ) && errorStr != NULL )
 				{
 				/* The last valid bit should be a one bit */
 				errorStr = "Spurious zero bits in bitstring";
@@ -1616,7 +1741,8 @@ static void dumpBitString( FILE *inFile, const int length, const int unused,
 			bitString <<= 1;
 			}
 		if( noBits < sizeof( int ) && \
-			( ( remainderMask << noBits ) & value ) )
+			( ( remainderMask << noBits ) & value ) && \
+			errorStr != NULL )
 			{
 			/* There shouldn't be any bits set after the last valid one.  We
 			   have to do the noBits check to avoid a fencepost error when
@@ -1655,7 +1781,7 @@ static void dumpBitString( FILE *inFile, const int length, const int unused,
 		printString( level, "%s", "'B\n" );
 
 	if( errorStr != NULL )
-		complain( errorStr, level );
+		complain( errorStr, 0, level );
 	}
 
 /* Display data as a text string up to a maximum of 240 characters (8 lines
@@ -1668,7 +1794,7 @@ static void dumpBitString( FILE *inFile, const int length, const int unused,
    it up into 48-char chunks in a somewhat less nice text-dump format */
 
 static void displayString( FILE *inFile, long length, int level,
-						   STR_OPTION strOption )
+						   const STR_OPTION strOption )
 	{
 	char timeStr[ 64 ];
 	long noBytes = length;
@@ -1852,13 +1978,13 @@ static void displayString( FILE *inFile, long length, int level,
 
 	/* Display any problems we encountered */
 	if( warnPrintable )
-		complain( "PrintableString contains illegal character(s)", level );
+		complain( "PrintableString contains illegal character(s)", 0, level );
 	if( warnIA5 )
-		complain( "IA5String contains illegal character(s)", level );
+		complain( "IA5String contains illegal character(s)", 0, level );
 	if( warnTime )
-		complain( "Time is encoded incorrectly", level );
+		complain( "Time is encoded incorrectly", 0, level );
 	if( warnBMP )
-		complain( "BMPString has missing final byte/half character", level );
+		complain( "BMPString has missing final byte/half character", 0, level );
 	}
 
 /****************************************************************************
@@ -1878,6 +2004,8 @@ static int getItem( FILE *inFile, ASN1_ITEM *item )
 	memset( item, 0, sizeof( ASN1_ITEM ) );
 	item->indefinite = FALSE;
 	tag = item->header[ index++ ] = fgetc( inFile );
+	if( tag == EOF )
+		return( FALSE );
 	item->id = tag & ~TAG_MASK;
 	tag &= TAG_MASK;
 	if( tag == TAG_MASK )
@@ -1910,9 +2038,12 @@ static int getItem( FILE *inFile, ASN1_ITEM *item )
 		}
 	fPos += 2;			/* Tag + length */
 	length = item->header[ index++ ] = fgetc( inFile );
+	if( length == EOF )
+		return( FALSE );
 	item->headerSize = index;
 	if( length & LEN_XTND )
 		{
+		const int lengthStart = index;
 		int i;
 
 		length &= LEN_MASK;
@@ -1929,11 +2060,34 @@ static int getItem( FILE *inFile, ASN1_ITEM *item )
 		for( i = 0; i < length; i++ )
 			{
 			int ch = fgetc( inFile );
-
+	
+			if( ch == EOF )
+				{
+				fPos += length - i;
+				return( FALSE );
+				}
 			item->length = ( item->length << 8 ) | ch;
 			item->header[ i + index ] = ch;
 			}
 		fPos += length;
+
+		/* Check for the length being less then 128, which means it 
+		   shouldn't be encoded as a long length */
+		if( item->length < 128 )
+			item->nonCanonical = lengthStart;
+
+		/* Check for the first 9 bits of the length being identical and
+		   if they are, remember where the encoded non-canonical length
+		   starts */
+		if( item->headerSize - lengthStart > 1 )
+			{
+			if( ( item->header[ lengthStart ] == 0x00 ) && \
+				( ( item->header[ lengthStart + 1 ] & 0x80 ) == 0x00 ) )
+				item->nonCanonical = lengthStart - 1;
+			if( ( item->header[ lengthStart ] == 0xFF ) && \
+				( ( item->header[ lengthStart + 1 ] & 0x80 ) == 0x80 ) )
+				item->nonCanonical = lengthStart - 1;
+			}
 		}
 	else
 		item->length = length;
@@ -2243,10 +2397,14 @@ static void printConstructed( FILE *inFile, int level, const ASN1_ITEM *item )
 	if( !item->length && !item->indefinite )
 		{
 		printString( level, "%s", " {}\n" );
+		if( item->nonCanonical )
+			complainLengthCanonical( item, level );
 		return;
 		}
 
 	printString( level, "%s", " {\n" );
+	if( item->nonCanonical )
+		complainLengthCanonical( item, level );
 	result = printAsn1( inFile, level + 1, item->length, item->indefinite );
 	if( result )
 		{
@@ -2267,7 +2425,7 @@ static void printASN1object( FILE *inFile, ASN1_ITEM *item, int level )
 	{
 	OIDINFO *oidInfo;
 	STR_OPTION stringType;
-	unsigned char buffer[ MAX_OID_SIZE ];
+	BYTE buffer[ MAX_OID_SIZE ];
 	const int nonOutlineObject = \
 			( doOutlineOnly && ( item->id & FORM_MASK ) != CONSTRUCTED ) ? \
 			TRUE : FALSE;
@@ -2302,7 +2460,9 @@ static void printASN1object( FILE *inFile, ASN1_ITEM *item, int level )
 		if( !item->length && !item->indefinite && !zeroLengthOK( item ) )
 			{
 			printString( level, "%c", '\n' );
-			complain( "Object has zero length", level );
+			complain( "Object has zero length", 0, level );
+			if( item->nonCanonical )
+				complainLengthCanonical( item, level );
 			return;
 			}
 
@@ -2320,6 +2480,8 @@ static void printASN1object( FILE *inFile, ASN1_ITEM *item, int level )
 		if( nonOutlineObject )
 			{
 			dumpHex( inFile, item->length, 1000, FALSE );
+			if( item->nonCanonical )
+				complainLengthCanonical( item, level );
 			printString( level, "%c", '\n' );
 			return;
 			}
@@ -2331,11 +2493,15 @@ static void printASN1object( FILE *inFile, ASN1_ITEM *item, int level )
 			{
 			/* It looks like a text string, dump it as text */
 			displayString( inFile, item->length, level, stringType );
+			if( item->nonCanonical )
+				complainLengthCanonical( item, level );
 			return;
 			}
 
 		/* This could be anything, dump it as hex data */
 		dumpHex( inFile, item->length, level, FALSE );
+		if( item->nonCanonical )
+			complainLengthCanonical( item, level );
 
 		return;
 		}
@@ -2377,7 +2543,9 @@ static void printASN1object( FILE *inFile, ASN1_ITEM *item, int level )
 	if( !item->length && !zeroLengthOK( item ) )
 		{
 		printString( level, "%c", '\n' );
-		complain( "Object has zero length", level );
+		complain( "Object has zero length", 0, level );
+		if( item->nonCanonical )
+			complainLengthCanonical( item, level );
 		return;
 		}
 	switch( item->tag )
@@ -2386,10 +2554,17 @@ static void printASN1object( FILE *inFile, ASN1_ITEM *item, int level )
 			{
 			int ch;
 
+			if( item->length != 1 )	
+				complainLength( item, level );
 			ch = getc( inFile );
 			printString( level, " %s\n", ch ? "TRUE" : "FALSE" );
 			if( ch != 0 && ch != 0xFF )
-				complain( "BOOLEAN has non-DER encoding", level );
+				{
+				complain( "BOOLEAN '%02X' has non-DER encoding", ch, 
+						  level );
+				}
+			if( item->nonCanonical )
+				complainLengthCanonical( item, level );
 			fPos++;
 			break;
 			}
@@ -2397,15 +2572,25 @@ static void printASN1object( FILE *inFile, ASN1_ITEM *item, int level )
 		case INTEGER:
 		case ENUMERATED:
 			if( item->length > 4 )
+				{
 				dumpHex( inFile, item->length, level, TRUE );
+				if( item->nonCanonical )
+					complainLengthCanonical( item, level );
+				}
 			else
+				{
 				printValue( inFile, item->length, level );
+				if( item->nonCanonical )
+					complainLengthCanonical( item, level );
+				}
 			break;
 
 		case BITSTRING:
 			{
 			int ch;
 
+			if( item->length < 2 )	
+				complainLength( item, level );
 			if( ( ch = getc( inFile ) ) != 0 )
 				{
 				printString( level, " %d unused bit%s",
@@ -2415,7 +2600,9 @@ static void printASN1object( FILE *inFile, ASN1_ITEM *item, int level )
 			if( !--item->length && !ch )
 				{
 				printString( level, "%c", '\n' );
-				complain( "Object has zero length", level );
+				complain( "Object has zero length", 0, level );
+				if( item->nonCanonical )
+					complainLengthCanonical( item, level );
 				return;
 				}
 			if( item->length <= sizeof( int ) )
@@ -2423,6 +2610,8 @@ static void printASN1object( FILE *inFile, ASN1_ITEM *item, int level )
 				/* It's short enough to be a bit flag, dump it as a sequence
 				   of bits */
 				dumpBitString( inFile, ( int ) item->length, ch, level );
+				if( item->nonCanonical )
+					complainLengthCanonical( item, level );
 				break;
 				}
 			/* Drop through to dump it as an octet string */
@@ -2449,9 +2638,13 @@ static void printASN1object( FILE *inFile, ASN1_ITEM *item, int level )
 					( !checkCharset && ( stringType == STR_IA5 || \
 										 stringType == STR_PRINTABLE ) ) ? \
 					STR_NONE : stringType );
+				if( item->nonCanonical )
+					complainLengthCanonical( item, level );
 				return;
 				}
 			dumpHex( inFile, item->length, level, FALSE );
+			if( item->nonCanonical )
+				complainLengthCanonical( item, level );
 			break;
 
 		case OID:
@@ -2460,6 +2653,8 @@ static void printASN1object( FILE *inFile, ASN1_ITEM *item, int level )
 			int length, isValid;
 
 			/* Hierarchical Object Identifier */
+			if( item->length < 3 )	
+				complainLength( item, level );
 			if( item->length > MAX_OID_SIZE )
 				{
 				fflush( stdout );
@@ -2498,7 +2693,9 @@ static void printASN1object( FILE *inFile, ASN1_ITEM *item, int level )
 					printString( level, "(%s)\n", oidInfo->comment );
 					}
 				if( !isValid )
-					complain( "OID has invalid encoding", level );
+					complain( "OID has invalid encoding", 0, level );
+				if( item->nonCanonical )
+					complainLengthCanonical( item, level );
 
 				/* If there's a warning associated with this OID, remember
 				   that there was a problem */
@@ -2509,16 +2706,21 @@ static void printASN1object( FILE *inFile, ASN1_ITEM *item, int level )
 				}
 
 			/* Print the OID as a text string */
-			isValid = oidToString( textOID, &length, buffer, ( int ) item->length );
+			isValid = oidToString( textOID, &length, buffer, 
+								   ( int ) item->length );
 			printString( level, " '%s'\n", textOID );
 			if( !isValid )
-				complain( "OID has invalid encoding", level );
+				complain( "OID has invalid encoding", 0, level );
+			if( item->nonCanonical )
+				complainLengthCanonical( item, level );
 			break;
 			}
 
 		case EOC:
 		case NULLTAG:
 			printString( level, "%c", '\n' );
+			if( item->nonCanonical )
+				complainLengthCanonical( item, level );
 			break;
 
 		case OBJDESCRIPTOR:
@@ -2530,24 +2732,48 @@ static void printASN1object( FILE *inFile, ASN1_ITEM *item, int level )
 		case VIDEOTEXSTRING:
 		case UTF8STRING:
 			displayString( inFile, item->length, level, STR_NONE );
+			if( item->nonCanonical )
+				complainLengthCanonical( item, level );
 			break;
 		case PRINTABLESTRING:
 			displayString( inFile, item->length, level, STR_PRINTABLE );
+			if( item->nonCanonical )
+				complainLengthCanonical( item, level );
 			break;
 		case BMPSTRING:
 			displayString( inFile, item->length, level, STR_BMP );
+			if( item->nonCanonical )
+				complainLengthCanonical( item, level );
 			break;
 		case UTCTIME:
 			displayString( inFile, item->length, level, STR_UTCTIME );
+			if( item->nonCanonical )
+				complainLengthCanonical( item, level );
 			break;
 		case GENERALIZEDTIME:
 			displayString( inFile, item->length, level, STR_GENERALIZED );
+			if( item->nonCanonical )
+				complainLengthCanonical( item, level );
 			break;
 		case IA5STRING:
 			displayString( inFile, item->length, level, STR_IA5 );
+			if( item->nonCanonical )
+				complainLengthCanonical( item, level );
 			break;
 		case T61STRING:
 			displayString( inFile, item->length, level, STR_LATIN1 );
+			if( item->nonCanonical )
+				complainLengthCanonical( item, level );
+			break;
+
+		case SEQUENCE:
+			printString( level, "%c", '\n' );
+			complain( "SEQUENCE has invalid primitive encoding", 0, level );
+			break;
+
+		case SET:
+			printString( level, "%c", '\n' );
+			complain( "SET has invalid primitive encoding", 0, level );
 			break;
 
 		default:
@@ -2555,8 +2781,11 @@ static void printASN1object( FILE *inFile, ASN1_ITEM *item, int level )
 			if( !doPure )
 				printString( level, "%s", INDENT_STRING );
 			doIndent( level + 1 );
-			printString( level, "%s", "Unrecognised primitive, hex value is:");
+			printString( level, "%s", 
+						 "Unrecognised primitive, hex value is:");
 			dumpHex( inFile, item->length, level, FALSE );
+			if( item->nonCanonical )
+				complainLengthCanonical( item, level );
 			noErrors++;		/* Treat it as an error */
 		}
 	}
@@ -2638,7 +2867,7 @@ static int printAsn1( FILE *inFile, const int level, long length,
 			{
 			seenEOC = TRUE;
 			if( !isIndefinite)
-				complain( "Spurious EOC in definite-length item", level );
+				complain( "Spurious EOC in definite-length item", 0, level );
 			}
 		if( !doPure && !nonOutlineObject )
 			{
@@ -3045,7 +3274,7 @@ int main( int argc, char *argv[] )
 	printAsn1( inFile, 0, LENGTH_MAGIC, 0 );
 	if( !useStdin && offset == 0 )
 		{
-		unsigned char buffer[ 16 ];
+		BYTE buffer[ 16 ];
 		long position = ftell( inFile );
 
 		/* If we're dumping a standalone ASN.1 object and there's further
